/* You can modify the values below to alter the default look of trywhy3 */


/*** TASKS ***/

/* Color of the icon used to show a task is being proved */
.why3-task-pending {
    color: blue;
}

/* Color of the icon used to show a task is proved */
.why3-task-valid {
    color:green;
}

/* Color of the icon used to show the status of a task is unproven */
.why3-task-unknown {
    color:orange;
}

/* Style used to for selected tasks */
.why3-task-selected {
    background:rgba(250,220,90,0.7);
}

/*** EDITOR ***/

/*** Warning: for highlighting properties, do not
    modify the position or z-index property.
    You should not use alpha-transparency (either opacity < 1 or rgba colors)
    since some highlight region overlaps. 
***/
#why3-editor {
    font-size: large;
}

#why3-task-viewer {
    font-size: large;
}

/* Color used to highlight errors in the editor */
.why3-error {
    background:rgb(231,113,116);
}

/* Used for messages written in the right pane that are not errors */
.why3-msg {
}

/* Color used to highlight negative premises of a task in the editor */
.why3-loc-neg-premise {
    background:rgb(255,128,64);
}

/* Color used to highlight goals of a task in the editor */
.why3-loc-goal {
    background:rgb(64,191,255);
}

/* Color used to highlight premises of a task in the editor */
.why3-loc-premise {
    background:rgb(100,233,134);
}

/* Color of the interface */
.why3-widget {
    background: #eee;
    color: #666;
    border-color: #888;
}

/* Used for insactive widgets and tabs */
.why3-widget.why3-inactive {
    background:#eee;
    border-color:#eee;
    color:#aaa;
}
/* Style used for buttons */
.why3-button {
    background-image: linear-gradient(to bottom, #eee 10%, #ccc);
    border-radius:3pt;
    border: solid 1pt #aaa;
}
/* Text color used for icons and buttons */
.why3-button,
.why3-icon  {
    color:#333;
}

/* Highliting used for context menu */
.why3-contextmenu li:hover,
.why3-contextmenu li:hover .why3-icon
{
    color: #eee;
    background: #222;
}