(* This file is generated by Why3's Coq driver *)
(* Beware! Only edit allowed sections below    *)
Require Import ZArith.
Require Import Rbase.
Inductive t (a:Type) :=
  | A : a -> t a
  | B : t a.
Implicit Arguments A.
Set Contextual Implicit.
Implicit Arguments B.
Unset Contextual Implicit.

Parameter f: unit -> unit.


(* YOU MAY EDIT THE CONTEXT BELOW *)

(* DO NOT EDIT BELOW *)

Theorem x : ~ ((A tt) = (B :(t unit))).
(* YOU MAY EDIT THE PROOF BELOW *)
discriminate.
Qed.
(* DO NOT EDIT BELOW *)


