[tools fast]
prover = "alt-ergo"
prover = "cvc3"
prover = "z3"
prover = "simplify"
prover = "spass"
prover = "yices"
prover = "eprover"
timelimit = 1

[tools array]
command = "why3-cpulimit %t %m -s z3 -smt %f 2>&1"
driver = "/usr/local/share/why3/drivers/z3_array.drv"
#loadpath = "examples/programs"
#use = "meta.Partial_incomplete"
timelimit = 1

[tools instantiate]
prover = "cvc3"
prover = "z3"
prover = "spass"
prover = "yices"
prover = "eprover"
timelimit = 1
loadpath = "examples/programs"
use = "meta.Partial"

[probs funny]
file = "talk290.mlw"
file = "course.mlw"
file = "vstte10_aqueue.mlw"
file = "isqrt.mlw"
file = "insertion_sort_list.mlw"
file = "vacid_0_sparse_array.mlw"
file = "vacid_0_red_black_trees_harness.mlw"
file = "bresenham.mlw"
transform = "split_goal"

[bench fast_and_funny]
tools = fast
tools = array
tools = instantiate
probs = funny
timeline = "encodebench.time"
average = "encodebench.avg"
csv = "encodebench.csv"
