[tools fast]
prover = "alt-ergo"
prover = "cvc3"
prover = "spass"
timelimit = 5


[probs funny]
file = "talk290.mlw"
transform = "split_goal"

[bench fast_and_funny]
tools = fast
probs = funny
timeline = "prgbench.time"
average = "prgbench.avg"
csv = "prgbench.csv"
